import sys, os
from distance_matrix_fast import *
from test_common import *
import numpy as np
import pandas as pd
from tqdm import tqdm
import argparse
import subprocess


extra_opts = ''
outfile = '/dev/null'
static_test_cases = ['SampleInput/randomwalk1K_nan.txt', 'SampleInput/poorly_conditioned_test.txt']
window_sizes_to_test = [5, 7, 13, 25, 33, 78, 92, 100, 189, 339, 500]
tile_sizes_to_test = [1024, 2048, 4096]
input_sizes_to_test = [20, 100, 250, 500, 1500, 8000]
matrix_sizes_to_test = [3, 10, 101]
thresholds_to_test = [0.0,0.125,0.5]

np.set_printoptions(edgeitems=30, linewidth=100000)
np.random.seed(13)

parser = argparse.ArgumentParser()
parser.add_argument('--executable', help='SCAMP executable to test, input \'pyscamp\' to use pyscamp', required=True)
parser.add_argument('--force_gpu', type=bool, help='Forces GPU-specific tests to run')
parser.add_argument('--output_file', help='File for test std output')
parser.add_argument('--extra_args', help='Extra arguments to be passed to each test invocation')
parser.add_argument('--window_sizes', type=int, nargs='*')
parser.add_argument('--tile_sizes', type=int, nargs='*')
parser.add_argument('--input_sizes', type=int, nargs='*')
parser.add_argument('--matrix_sizes', type=int, nargs='*')
parser.add_argument('--thresholds', type=float, nargs='*')
args = parser.parse_args()


executable = args.executable

if args.output_file is not None:
  outfile = args.output_file

if args.extra_args is not None:
  extra_opts = args.extra_args

if args.window_sizes is not None:
  window_sizes_to_test = args.window_sizes

if args.tile_sizes is not None:
  tile_sizes_to_test = args.tile_sizes

if args.input_sizes is not None:
  input_sizes_to_test = args.input_sizes

if args.matrix_sizes is not None:
  matrix_sizes_to_test = args.matrix_sizes

if args.thresholds is not None:
  thresholds_to_test = args.thresholds

if executable == 'pyscamp':
  import pyscamp as mp
  if args.tile_sizes:
    print('Warning: ignoring tile_sizes during pyscamp execution')
  tile_sizes_to_test = [0]

gpu_enabled = False
if args.force_gpu or (executable == 'pyscamp' and mp.gpu_supported()):
  gpu_enabled = True

index_match_ratio = 0.001
matrix_match_ratio = 0.001

vector_match_epsilon_SUM = 0.001
vector_match_epsilon_1NN = 0.001

matrix_match_epsilon = 0.001

def generate_input_arrays(input_sizes):
  arrs = []
  for s in input_sizes:
    arrs.append(np.random.randn(s))
  return arrs

def read_file_inputs(files):
  arrs = []
  for fpath in files:
    x = np.array(pd.read_csv(fpath, sep=' ', header=None))
    arrs.append(x.reshape((len(x),)))
  return arrs

def generate_tests(input_sizes, windows):
  tests = set()
  for a in range(len(input_sizes)):
    for b in range(len(input_sizes)):
        exclusion = 0
        if a == b: 
          exclusion = input_sizes[a] // 4
        for window in windows:
          if window <= input_sizes[a] - exclusion and window <= input_sizes[b] - exclusion:
            tests.add((a, b, window))
  return tests


def evaluate_result(dm_reductions, scamp_results, subtestargs):
  ptype = subtestargs['ptype']
  if ptype == "1NN_INDEX":
    valid_nn, valid_idx = dm_reductions[("1NN_INDEX",None,None,None)]
    return compare_index(valid_idx, valid_nn, scamp_results[1], scamp_results[0]) and compare_vectors(valid_nn, scamp_results[0])
  
  if ptype == "1NN":
    valid_data = dm_reductions[("1NN",None,None,None)]
    return compare_vectors(valid_data, scamp_results[0])
    
  if ptype == "SUM_THRESH":
    valid_data = dm_reductions[("SUM_THRESH",subtestargs['threshold'],None,None)]
    return compare_vectors_sum(valid_data, scamp_results[0], subtestargs['threshold'])
  
  if ptype == "MATRIX_SUMMARY":
    valid_data = dm_reductions[("MATRIX_SUMMARY", None, subtestargs['rrow'], subtestargs['rcol'])] 
    return compare_matrix(valid_data, scamp_results[0], 0.0)
  
  if ptype == "ALL_NEIGHBORS":
    valid_dm = dm_reductions[("ALL_NEIGHBORS", None, None, None)]
    valid_nn, valid_idx = dm_reductions[("1NN_INDEX",None,None,None)]
    return compare_all_neighbors(valid_dm, valid_nn, valid_idx, scamp_results[0], subtestargs['threshold'])
  
  return None

def run_pyscamp(inputs, a, b, window, max_matches, thresh, ptype, rrows, rcols):  
  args = {}
  args['pearson'] = True
  if thresh:
    args['threshold'] = thresh
  if rrows:
    args['mheight'] = rrows
  if rcols:
    args['mwidth'] = rcols
  if '--no_gpu' in extra_opts:
    args['gpus'] = []

  if not max_matches:
    max_matches = 5
  
  mp_columns_out = None
  mp_columns_out_index = None
  mp_rows_out = None
  mp_rows_out_index = None
    
  if ptype == "1NN_INDEX":
    if a == b:
      mp_columns_out, mp_columns_out_index = mp.selfjoin(inputs[a], window, **args)
    else:
      mp_columns_out, mp_columns_out_index = mp.abjoin(inputs[a], inputs[b], window, **args)
  elif ptype == "SUM_THRESH":
    if a == b:
      mp_columns_out = mp.selfjoin_sum(inputs[a], window, **args)
    else:
      mp_columns_out = mp.abjoin_sum(inputs[a], inputs[b], window, **args)
  elif ptype == "ALL_NEIGHBORS":
    if a == b:
      mp_columns_out = mp.selfjoin_knn(inputs[a], window, max_matches, **args)
    else:
      mp_columns_out = mp.abjoin_knn(inputs[a], inputs[b], window, max_matches, **args)
  elif ptype == "MATRIX_SUMMARY":
    if a == b:
      mp_columns_out = mp.selfjoin_matrix(inputs[a], window, **args)
    else:
      mp_columns_out = mp.abjoin_matrix(inputs[a], inputs[b], window, **args)
  else:
    raise ValueError('pyscamp does not support profile type {}'.format(ptype))

  if mp_columns_out is not None:
    mp_columns_out = mp_columns_out.squeeze()
  if mp_columns_out_index is not None:
    mp_columns_out_index = mp_columns_out_index.squeeze()
  if mp_rows_out is not None:
    mp_rows_out = mp_rows_out.squeeze()
  if mp_rows_out_index is not None:
    mp_rows_out_index = mp_rows_out_index.squeeze()
  
  return mp_columns_out, mp_columns_out_index, mp_rows_out, mp_rows_out_index

def read_file_to_array(filename):
  if not os.path.exists(filename):
    return None
  try:
    return np.array(pd.read_csv(filename, sep=' ', header=None, na_values='-nan(ind)')).squeeze()
  except pd.errors.EmptyDataError:
    print('Note: file {} was empty. Returning empty array.'.format(filename))
    return np.array([])

def run_scamp(inputs, a, b, window, tilesz, max_matches, thresh, ptype, rrows, rcols, keep_rows, aligned):
  if executable == 'pyscamp':
    if keep_rows:
      raise ValueError('keep_rows not supported with pyscamp.')
    if aligned:
      raise ValueError('aligned not supported with pyscamp.')
    return run_pyscamp(inputs, a, b, window, max_matches, thresh, ptype, rrows, rcols)

  args = f'--output_pearson --window={window} --input_a_file_name=a.txt {extra_opts}'

  if a != b:
    args += f' --input_b_file_name=b.txt'
  args += f' --max_tile_size={tilesz} --profile_type={ptype}'

  if keep_rows:
    args += ' --keep_rows_separate'

  if aligned:
    args += ' --aligned'

  if thresh is not None:
    args += f' --threshold={thresh}'
  
  if rrows is not None and rcols is not None:
    args += f' --reduced_height={rrows} --reduced_width={rcols}'

  if max_matches is not None:
    args += f' --max_matches_per_column={max_matches}'

  print(args)
  
  ret = subprocess.call(os.path.abspath(executable) + ' ' + args, shell=True)
  
  mp_columns_out = read_file_to_array('mp_columns_out')
  mp_columns_out_index = read_file_to_array('mp_columns_out_index')  
  mp_rows_out = read_file_to_array('mp_rows_out')
  mp_rows_out_index = read_file_to_array('mp_rows_out_index')

  return mp_columns_out, mp_columns_out_index, mp_rows_out, mp_rows_out_index


def run_test(test, inputs):
  a = test[0]
  a_data = inputs[a]
  b = test[1]
  b_data = inputs[b]
  window = test[2]
  if a != b:
    dm = distance_matrix_np(a_data,b_data,window)
  else:
    dm = distance_matrix_np(a_data,None,window)

  dm_reductions = {}
  dm_reductions[('1NN_INDEX', None, None, None)] = reduce_1nn_index(dm)
  dm_reductions[('1NN', None, None, None)] = reduce_1nn(dm)
  dm_reductions[('ALL_NEIGHBORS', None, None, None)] = dm
  
  result_sum = []
  for thresh in thresholds_to_test:
    dm_reductions[('SUM_THRESH', thresh, None, None)] = reduce_sum_thresh(dm, thresh)
  for rows in matrix_sizes_to_test:
    if dm.shape[0] < rows:
      continue
    for cols in matrix_sizes_to_test:
      if dm.shape[1] < cols:
        continue
      dm_reductions[('MATRIX_SUMMARY', None, rows, cols)] = reduce_matrix(dm, rows,cols, a == b)
  
  # We only need local files for the CLI
  if executable != 'pyscamp':
    np.savetxt('a.txt', a_data)
    np.savetxt('b.txt', b_data)
  
  subtests = {}
  prev_tile_size = None
  for tile_sz in tile_sizes_to_test:
    if prev_tile_size is not None and prev_tile_size > len(a_data) and prev_tile_size > len(b_data):
      break
    subtest_dict = {'tilesz' : tile_sz, 'matchpercol' : None, 'threshold': None, 'ptype': "1NN_INDEX", 'rrow': None, 'rcol': None, 'keeprows': False, 'aligned': False}
    subtest_args = tuple(subtest_dict.values())
    scamp_results = run_scamp(inputs, a, b, window, tile_sz, None, None, "1NN_INDEX", None, None,False, False);
    valid = evaluate_result(dm_reductions,scamp_results,subtest_dict)
    subtests[subtest_args] = valid
 
    # Pyscamp does not support 1NN profiles
    if executable != 'pyscamp':
      subtest_dict = {'tilesz' : tile_sz, 'matchpercol' : None, 'threshold': None, 'ptype': "1NN", 'rrow': None, 'rcol': None, 'keeprows': False, 'aligned': False}
      subtest_args = tuple(subtest_dict.values())
      scamp_results = run_scamp(inputs, a, b, window, tile_sz, None, None, "1NN", None, None,False, False);
      valid = evaluate_result(dm_reductions,scamp_results,subtest_dict)
      subtests[subtest_args] = valid
   

    for thresh in thresholds_to_test:
      subtest_dict = {'tilesz' : tile_sz, 'matchpercol' : None, 'threshold': thresh, 'ptype': "SUM_THRESH", 'rrow': None, 'rcol': None, 'keeprows': False, 'aligned': False}
      subtest_args = tuple(subtest_dict.values())
      scamp_results = run_scamp(inputs, a, b, window, tile_sz, None, thresh, "SUM_THRESH", None, None,False, False);
      valid = evaluate_result(dm_reductions,scamp_results,subtest_dict)
      subtests[subtest_args] = valid

      # KNN MPs are only supported when cuda devices are available and SCAMP is built with CUDA.
      # KNN not supported on both CPU/GPU yet
      '''
      if gpu_enabled: 
        subtest_dict = {'tilesz': tile_sz, 'matchpercol': 5, 'threshold': thresh, 'ptype': 'ALL_NEIGHBORS', 'rrow': None, 'rcol': None, 'keeprows': False, 'aligned': False}
        subtest_args = tuple(subtest_dict.values())
        scamp_results = run_scamp(inputs, a, b, window, tile_sz, 5, thresh, "ALL_NEIGHBORS", None, None,False, False);
        valid = evaluate_result(dm_reductions,scamp_results,subtest_dict)
        subtests[subtest_args] = valid
      '''

    for rrow in matrix_sizes_to_test:
      if rrow >= len(inputs[b]) - window + 1:
        continue
      for rcol in matrix_sizes_to_test:
        if rcol >= len(inputs[a]) - window + 1:
          continue
        # GPUs do not currently output the exact matrix summary, it is not currently possible to use the current verification method on GPU output.
        if not gpu_enabled:
          subtest_dict = {'tilesz' : tile_sz, 'matchpercol' : None, 'threshold': None, 'ptype': "MATRIX_SUMMARY", 'rrow': rrow, 'rcol': rcol, 'keeprows': False, 'aligned': False}
          subtest_args = tuple(subtest_dict.values())
          scamp_results = run_scamp(inputs, a, b, window, tile_sz, None, None, "MATRIX_SUMMARY", rrow, rcol, False, False);
          valid = evaluate_result(dm_reductions,scamp_results,subtest_dict)
          subtests[subtest_args] = valid

    prev_tile_size = tile_sz

  if os.path.exists('a.txt'):
    os.remove('a.txt')
  if os.path.exists('b.txt'):
    os.remove('b.txt')
        
  return subtests

def all_tests_passed(results):
  test_count = 0
  correct_tests = 0
  for t, subtests in results:
    for key, test in subtests.items():
      test_count += 1
      if test:
        correct_tests += 1
      else:
         print(t)
         print(key)
  print(f'{correct_tests} of {test_count} tests passed')
  return test_count == correct_tests

inputs = generate_input_arrays(input_sizes_to_test)
file_inputs = read_file_inputs(static_test_cases)
inputs += [x for x in file_inputs]

input_sizes_to_test += [len(x) for x in file_inputs]

tests = generate_tests(input_sizes_to_test, window_sizes_to_test)

all_results = []
for test in tqdm(tests):
  alen = len(inputs[test[0]])
  blen = len(inputs[test[1]])
  results = run_test(test, inputs)
  for key, result in results.items():
    if not result:
      print(f'A = {alen}, B = {blen}, m = {test[2]}, tile size = {key[0]}, matchpercol = {key[1]}, threshold = {key[2]}, profile type = {key[3]}, reduced_rows = {key[4]}, reduced cols = {key[5]}, keep_rows = {key[6]}, aligned = {key[7]}')
      print('FAIL!')
  all_results.append([test,results])


if all_tests_passed(all_results):
  exit(0)

exit(1)




#!/usr/bin/python
import os
import sys
import math

fd1 = open(sys.argv[1], 'r')
fd2 = open(sys.argv[2], 'r')
fd3 = open(sys.argv[3], 'w')

maxdiff = 0
count = 1
for line1, line2 in zip(fd1,fd2):
    if 'nan' in line1:
      line1 = 'nan' 
    if 'nan' in line2:
      line2 = 'nan'
    x = float(line1)
    x2 = float(line2)
    if math.isnan(x) != math.isnan(x2) or math.isinf(x) != math.isinf(x2):
      print('Failure: values do not agree') 
      print('Line: ', count, x, x2)
      exit(1)
    diff = x - x2;
    fd3.write(str(x - x2) + "\n")
    if(abs(diff) > maxdiff):
        maxdiff = abs(diff)
    if abs(diff) > 0.01:
        print('Failure: matrix profile differs from the ground truth by more than the threshold')  
        print('Line: ', count, x, x2)
        exit(1)
    count += 1

print("Max matrix profile difference was " + str(maxdiff))






import os
import sys
import random


#Generates a random list of floating point values between 0 and 10
fd = open(sys.argv[3], "w")
upper_lower_bound = float(sys.argv[2]);
num = 0
for i in range(0, int(sys.argv[1])):
    num += random.uniform(-upper_lower_bound,upper_lower_bound);
    fd.write(str(num) + "\n")

fd.close()



import numpy as np

def compare_index(valid, valid_corr, check, check_corr):
  if np.any(valid.shape != check.shape):
    print('Compare Index: Output Shapes valid = {} and check = {} do not match'.format(valid.shape, check.shape))
    return False

  # Find the elements which do not agree
  no_agree = valid != check

  # SCAMP performs the best so far comparison in FP32, so if the correlations are FP32 equivalent this is okay.
  corr_fp32_no_agree = valid_corr[no_agree].astype(np.float32) != check_corr[no_agree].astype(np.float32)

  # Any indexes whose nearest neighbor correlations do not agree with fp32 precision should be treated as non-matching
  incorrect = np.count_nonzero(corr_fp32_no_agree)

  if incorrect > 0:
    print('Compare Index: found {} incorrect indexes.'.format(incorrect))
    return False
  return True


def compare_vectors(valid, check):
  if np.any(valid.shape != check.shape):
    print('Compare Vectors: Output Shapes valid = {} and check = {} do not match'.format(valid.shape, check.shape))
    return False
  if not np.allclose(valid, check, equal_nan=True):
    print('Compare Vectors: Found inconsistencies between SCAMP output and the ground truth.')
    return False
  return True

def compare_vectors_sum(valid, check, thresh):
  if np.any(valid.shape != check.shape):
    print('Compare Sum: Output Shapes valid = {} and check = {} do not match'.format(valid.shape, check.shape))
    return False
    
  x = np.allclose(valid, check, equal_nan=False)
  if not x:
    # Check if the errors are all very close to the threshold
    # Find the number of times we can fit the threshold into the difference. In the case of slight perturbation errors on results around the threshold, these should be very close to whole numbers
    diff = abs(valid - check) / thresh

    # Round to the nearest 6th decimal place (the values could be very slightly over or under the nearest whole number, we want to make sure we treat these close values as whole numbers)
    diff = np.round(diff, decimals=6)

    # Subtract the whole number part, leaving only the decimal
    diff = diff - np.floor(diff)

    # Ignore any perturbations smaller than the 6th decimal place
    diff[diff < 1e-6] = 0

    # The nonzero values in this resulting array correspond to errors that very likely do not come from perturbation errors around the threshold.
    incorrect = np.count_nonzero(diff)
    
    if incorrect != 0:
      print('Compare Sum: Found {} values which were far from the correct sum.'.format(incorrect))
      return False
  return True

def compare_matrix(valid_matrix, check, thresh):
  if np.any(valid_matrix.shape != check.shape):
    print('Compare Matrix: Output Shapes valid = {} and check = {} do not match'.format(valid_matrix.shape, check.shape))
    return False

  if thresh is not None:
    valid = np.copy(valid_matrix)
    valid[valid < thresh] = np.nan
  else:
    valid = valid_matrix
  
  return np.allclose(valid, check, equal_nan=True)

# This only verifies that the outputs generated by ALL_NEIGHBORS are valid 
# entries that match the true distance matrix. It does not validate that
# we found all the matches we should have found or a match we weren't supposed to find
def compare_all_neighbors(valid_matrix, valid_nearest, valid_nearest_index, check, thresh):
  rows = [int(x[1]) for x in check]
  cols = [int(x[0])for x in check]
  corrs = [x[2] for x in check]

  # All results should be above the threshold
  if np.any(np.less(corrs,thresh)):
    print('Not all results were above the threshold')
    return False

  # Duplicates should not appear.
  prev_col = -1
  row_set = set()
  nearest = np.ones(len(valid_nearest)) * -1
  nearest_index = np.ones(len(valid_nearest_index)) * -1
  for col, row, corr in zip(cols, rows, corrs):
    if col != prev_col:
      row_set = set()
      prev_col = col
      nearest[col] = corr
      nearest_index[col] = row
    if row in row_set:
      print('Duplicate row {} found in col {}'.format(row, col))
      return False
    else:
      row_set.add(row)
  
  missing_idx = nearest == -1
  valid_missing = valid_nearest[missing_idx]
  valid_missing = valid_missing[~np.isnan(valid_missing)]
  # If there are results completely missing they should have a nearest neighbor less than the threshold.
  if not np.all(valid_missing < thresh):
    print('Missing indexes were not less than the threshold')
    return False

  if (len(rows) == 0):
    return True
  
  # All outputs are present in the distance matrix somewhere
  if not np.allclose(valid_matrix[rows, cols], corrs, equal_nan=True):
    print('Not all outputs were present in the distance matrix')
    return False
    
  # Indexes which are missing due to the matches being lower than the threshold should not count against us.
  nearest[missing_idx] = valid_nearest[missing_idx] 
  nearest_index[missing_idx] = valid_nearest_index[missing_idx] 
  
  # Nearest neighbor should appear for each column. We can only guarentee the nearest neighbor will appear
  # the 2nd, 3rd ... neighbors may be different depending on how the reduction is done.
  if not compare_vectors(valid_nearest, nearest):
    printf('Nearest neighbors did not match.')
    return False 
  if not compare_index(valid_nearest_index, valid_nearest, nearest_index, nearest):
    printf('Nearest neighbor indexes did not match.')
    return False 
  return True    



import numpy as np
import math

def moving_mean(a, w):
  result = np.zeros((len(a) - w + 1,))
  p = a[0]
  s = 0
  for i in range(1, w):
    x = p + a[i]
    z = x - p
    s += (p - (x - z)) + (a[i] - z)
    p = x
  
  result[0] = (p + s) / w

  for i in range(w, len(a)):
    x = p - a[i - w]
    z = x - p
    s += (p - (x - z)) - (a[i - w] + z)
    p = x

    x = p + a[i]
    z = x - p
    s += (p - (x - z)) + (a[i] - z)
    p = x
    result[i - w + 1] = (p + s) / w

  return result;

def sum_of_squared_differences(a, means, w):
  result = np.zeros((len(a) - w + 1,))
  for i in range(len(a) - w + 1):
    vals = a[i:i+w] - means[i]
    vals = vals * vals
    result[i] = np.sum(vals)
  return result
  
def get_precomputes(T, m, nanvalues):
  flatness_epsilon = 1e-13
  n = len(T) - m + 1;
  df = np.zeros((n,))
  dg = np.zeros((n,))

  means = moving_mean(T,m)

  norms = sum_of_squared_differences(T, means, m)

  for i in range(n):
    if nanvalues[i]:
      norms[i] = np.nan
    elif norms[i] <= flatness_epsilon:
      norms[i] = np.nan
    else:
      norms[i] = 1.0 / math.sqrt(norms[i])

  for i in range(n-1):
    df[i] = (T[i + m] - T[i]) / 2.0;
    dg[i] = (T[i + m] - means[i + 1]) + (T[i] - means[i]);

  return means, norms, df, dg
    
def convert_non_finite_to_zero(T, m):
  timeseries_clean = np.zeros((len(T),))
  nanvals = np.zeros((len(T) - m + 1,))
  steps_since_last_nan = m
  for i in range(len(T)):
    if np.isfinite(T[i]):
      timeseries_clean[i] = T[i]
    else:
      steps_since_last_nan = 0
      timeseries_clean[i] = 0
    if i >= m - 1:
      nanvals[i - m + 1] = steps_since_last_nan < m
    steps_since_last_nan += 1
  return timeseries_clean, nanvals

# Computes the distance matrix using the diagonal update method used in SCAMP
def distance_matrix(a,b,w):
    has_b = True
    if b is None:
        has_b = False
        b = a
    na = len(a) - w + 1
    if not has_b:
        nb = na
    else:
        nb = len(b) - w + 1
    out = np.ones((nb,na)) * -2


    a, nan_a = convert_non_finite_to_zero(a,w)
  
    if has_b:
      b, nan_b = convert_non_finite_to_zero(b,w)
    else:
      b = a;

    mua, siga, dfa, dga = get_precomputes(a,w,nan_a)
    if not has_b:
        mub = mua
        sigb = siga
        dfb = dfa
        dgb = dga
    else:
        mub, sigb, dfb, dgb = get_precomputes(b,w, nan_b)

    diagmax = na
    if not has_b:
        minlag = w // 4
    else:
        minlag = 0

    
    c = np.zeros((diagmax - minlag,)) 
    for diag in range(minlag,diagmax):
      c[diag-minlag] = np.sum((a[diag:diag+w]-mua[diag]) * (b[:w]-mub[0]))
    
    for offset in range(nb-minlag):
      result = c*(sigb[offset]*siga[minlag+offset:])
      out[offset, minlag+offset:] = result
      if not has_b:
        out[minlag+offset:, offset] = result
      x = c + dfb[offset] * dga[minlag+offset:] + dfa[minlag+offset:]*dgb[offset]
      c = x[:-1]
    if has_b:
      diagmax = nb
      c = np.zeros((diagmax - minlag,)) 
      for diag in range(minlag,diagmax):
        c[diag-minlag] = np.sum((b[diag:diag+w]-mub[diag]) * (a[:w]-mua[0]))
      for offset in range(na-minlag):
        result = c*(siga[offset]*sigb[minlag+offset:])
        out[minlag+offset:, offset] = result
        x = c + dfa[offset] * dgb[minlag+offset:] + dfb[minlag+offset:]*dga[offset]
        c = x[:-1]

    out[np.isnan(out)] = -2;
    return out


# Computes the distance matrix using np.corrcoef
def distance_matrix_np(a,b,m):
  has_b = True
  if b is None:
    has_b = False
    b = a

  na = len(a) - m  + 1
  nb = len(b) - m  + 1

  a, nan_a = convert_non_finite_to_zero(a,m)

  if has_b:
    b, nan_b = convert_non_finite_to_zero(b,m)
  else:
    b = a;

  mua, siga, _, _ = get_precomputes(a,m,nan_a)
  if not has_b:
      mub = mua
      sigb = siga
  else:
      mub, sigb, _, _ = get_precomputes(b,m, nan_b)

  
  x = np.zeros((na,m))
  y = np.zeros((nb,m))
  for i in range(na):
    if (np.isnan(siga[i])):
      x[i,:] = np.nan
    else:
      x[i,:] = a[i:i+m]

  if has_b:
    for i in range(nb):
      if (np.isnan(sigb[i])):
        y[i,:] = np.nan
      else:
        y[i,:] = b[i:i+m]

  if has_b:
    out = np.corrcoef(x,y=y)
    # Take the AB-join portion of the correlation matrix
    out = out[na:, :na]
  else:
    out = np.corrcoef(x)

  # Mark exclusion zone
  if not has_b:
    minlag = m // 4 - 1
    for i in range(nb):
      x = max(0, i - minlag)
      y = min(na, i + minlag + 1)
      out[i,x:y] = -2 

  out[np.isnan(out)] = -2;
  return out

# Simple (slow) distance matrix generation using scipy.stats.pearsonr for reference
'''
def distance_matrix_simple(a,b, m):
  has_b = True
  if b is None:
    has_b = False
    b = a
  na = len(a) - m  + 1
  nb = len(b) - m  + 1
  out = np.ones((nb,na)) * -2
  for i in range(na):
    x = a[i:i+m]
    for j in range(nb):
      out[j,i] = pearsonr(x, b[j:j+m])[0]
  if not has_b:
    minlag = m // 4 - 1
    for i in range(nb):
      x = max(0, i - minlag)
      y = min(na, i + minlag + 1)
      out[i,x:y] = -2 
  out[np.isnan(out)] = -2;
  return out
'''
    
def reduce_1nn_index(dm):
    corr = np.amax(dm, axis=0)
    idxs = np.argmax(dm, axis=0)
    idxs[corr == -2] = -1
    corr[corr == -2] = np.nan
    return corr.squeeze(), idxs.squeeze()

def reduce_1nn(dm):
    corrs = np.amax(dm, axis=0)
    corrs[corrs == -2] = np.nan
    return corrs.squeeze()

def reduce_sum_thresh(dm, thresh):
    dm2 = np.copy(dm)
    dm2[dm2 <= thresh] = 0

    result = np.sum(dm2, dtype='float64', axis=0)
    return result.squeeze()

def reduce_frequency_thresh(dm, thresh):
    dm2 = np.copy(dm)
    dm2[dm2 > thresh] = 1
    dm2[dm2 <= thresh] = 0
  
    result = np.sum(dm2, dtype='float64', axis=0)
    return result.squeeze()
  

def reduce_matrix(dm_orig, rows, cols, self_join):
    dm = np.copy(dm_orig)
    if self_join:
      # In a self join. SCAMP only computes the upper triagnular
      # portion of the distance matrix. We need to erase the bottom
      # half to not get a different reduction in these cases. 
      for col in range(dm.shape[1]):
        if col + 1 >= dm.shape[0]:
          break
        dm[col + 1:, col] = np.nan

    reduced_rows = dm.shape[0] / rows
    reduced_cols = dm.shape[1] / cols
    out = np.ones((rows,cols)) * -2
    for r in range(rows):
      st_r = math.ceil(r * reduced_rows)
      ed_r = min(dm.shape[0], math.ceil((r+1)* reduced_rows))
      for c in range(cols):
        st_c = math.ceil(c * reduced_cols)
        ed_c = min(dm.shape[1], math.ceil((c+1)*reduced_cols))
        out[r, c] = np.nanmax(dm[st_r:ed_r, st_c:ed_c])
    out[out == -2] = np.nan
    return out




import os
import sys
import random


#Generates a random list of floating point values between 0 and 10
fd = open(sys.argv[2], "w")

for i in range(0, int(sys.argv[1])):
    fd.write(str(random.randint(0, 1000000000) / float(100000000)) + "\n")

fd.close()



#This file includes tests to check the correcntess of the pyscamp module.
#It does not do extenstive testing of the SCAMP framework that is handled by other integration tests.

import pyscamp as mp
import numpy as np
from distance_matrix_fast import *
from test_common import *
import random 
import sys

failed = False
arr = []
arr2 = []

arr = np.random.random(size=(8000,))
arr2 = np.random.random(size=(8000,))

dm_self = distance_matrix(arr, None, 1024)
dm_ab = distance_matrix(arr, arr2, 1024)

dist, index = mp.selfjoin(arr, 1024, pearson=True)
vdist, vindex = reduce_1nn_index(dm_self)
if compare_vectors(vdist, dist) and compare_index(vindex, vdist, index, dist):
  print("1NN INDEX Self join pass")
else:
  failed = True
  print("1NN INDEX Self join fail")


dist, index = mp.abjoin(arr, arr2, 1024, pearson=True)
vdist, vindex = reduce_1nn_index(dm_ab)


if compare_vectors(vdist, dist) and compare_index(vindex, vdist, index, dist):
  print("1NN INDEX AB join pass")
else:
  failed = True
  print("1NN INDEX AB join fail")

dist = mp.selfjoin_sum(arr, 1024, threshold=0.90, pearson=True)
vdist = reduce_sum_thresh(dm_self, 0.90)

if compare_vectors_sum(vdist, dist, 0.90):
  print("SUM Self join pass")
else:
  failed = True
  print("SUM Self join fail")


dist = mp.abjoin_sum(arr, arr2, 1024, threshold=0.90, pearson=True)
vdist = reduce_sum_thresh(dm_ab, 0.90)

if compare_vectors_sum(vdist, dist, 0.90):
  print("SUM AB join pass")
else:
  failed = True
  print("SUM AB join fail")

dist = mp.selfjoin(arr, 1024, threads=1)
dist = mp.selfjoin(arr, 1024, threads=2)

matrix_summary = reduce_matrix(dm_self, 5, 10, True)
matrix_out = mp.selfjoin_matrix(arr, 1024, threshold=0.05, mwidth=10, mheight=5, pearson=True)
if compare_matrix(matrix_summary, matrix_out, 0.05):
  print ("Matrix Summary self join pass")
else:
  failed = True
  print ("Matrix Summary self join fail")

matrix_summary = reduce_matrix(dm_ab, 5, 10, False)
matrix_out = mp.abjoin_matrix(arr, arr2, 1024, threshold=0.05, mwidth=10, mheight=5, pearson=True)
if compare_matrix(matrix_summary, matrix_out, 0.05):
  print ("Matrix Summary AB join pass")
else:
  failed = True
  print ("Matrix Summary AB join fail")

if mp.gpu_supported():
  print('GPUs Supported')
  thresh = 0.12
  vdist, vindex = reduce_1nn_index(dm_self)
  x = mp.selfjoin_knn(arr, 1024, 5, threshold=thresh, pearson=True)
  if compare_all_neighbors(dm_self, vdist, vindex, x, thresh):
    print("KNN Self join pass")
  else:
    failed = True
    print("KNN Self join fail")
  vdist, vindex = reduce_1nn_index(dm_ab)
  x = mp.abjoin_knn(arr,arr2, 1024, 5, threshold=thresh, pearson=True)
  if compare_all_neighbors(dm_ab, vdist, vindex, x, thresh):
    print("KNN AB join pass")
  else:
    failed = True
    print("KNN AB join fail")

if failed:
  exit(1)


